//-----------------------------------------------------------------------------------------
// StrFind : General purpose string find and replace utility.
//
//  Copyright (c) 2004, Andrew J. Gryc
//
//	http://gryc.ws
//	andy@gryc.ws
//
// This program is free software; you can redistribute it and/or modify   
// it under the terms of the GNU General Public License as published by   
// the Free Software Foundation; either version 2 of the License, or (at  
// your option) any later version.                                        
//                                                                        
// This program is distributed in the hope that it will be useful, but    
// WITHOUT ANY WARRANTY; without even the implied warranty of             
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU      
// General Public License for more details.                              
//                                                                        
// For a copy of the GNU General Public License, write to the Free        
// Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA     
// 02111-1307, USA.                                                       
//-----------------------------------------------------------------------------------------
#include "stdafx.h"

// Because template classes have very long "real" names, suppress warnings about them
// exceeding length.  Unfortunately, MSVC++ has a bug that prevents this warning disable
// from always working, so just try to ignore the errors anyway!
#pragma warning(disable : 4786)

// Visual Studio .NET is very annoying about warning deprecation of tons of standard C stuff.
// Turn this off--we're not going to get forced into supporting only what Microsoft wants...
#pragma warning(disable : 4996)

#include <vector>
#include <algorithm>
#include <stdio.h>
#include <windows.h>
#include <iostream>
#include <fstream>
#include <ios>
#include <string>
#include <iomanip>

using namespace std;

// Our case-insensitive string
#include "CIString.h"

#include "input_line.h"
#include "IndexFile.h"

// Just in case we ever convert to POSIX compliant someday
#ifdef WIN32
const char DIR_SEPARATOR = '\\';
const char *NULL_DEVICE = "nul";
#else
const char DIR_SEPARATOR = '/';
const char *NULL_DEVICE = "/dev/nul";
#endif

// Length of our progress line
#define MAX_DISPLAYED_CHARS 75

// Match tag we print before each file that matches
#define MATCHHEADER         ">> File : "
#define MATCHSTRING			">> Within "

// Default name of index file
#define INDEX_FILE          ".strfind.index"

IndexFile *idx          = 0;


// All options and their defaults
bool PerformReplace		= false;
bool CaseInsensitive	= false;
bool SameCapsForm		= false;
bool InvertedMatch		= false;
bool MatchWholeLine		= false;
bool ProcessEscapes		= true;

bool PrintTotalStats	= true;
bool PrintGroupStats	= true;
bool PrintMatches		= true;
bool PrintBeforeMatch	= false;
bool PrintLinenumbers	= false;
bool PrintFilenames		= false;

bool CreateBackups		= false;
bool Subdirectories		= false;
bool ReturnErrorCount	= false;
bool BinaryMode		    = false;
bool FullyEnumerateDirs	= false;
bool PrintProgress		= true;
bool UnicodeMode		= false;
bool QuickIndex         = false;
int  Verbosity          = 0;

char indexfilename[MAX_PATH];

// Are we showing a progress line currently?
bool showing_progress;

void PrintHelp()
{
	// We've consumed all letters as options; we'll either have to get more creative,
    // use upper and lower case, or start naming them.

	cout << "StrFind: [<options>] <search string> [<replace string>] [<file wildcard>]+" << endl;
	cout << "Version 1.5  Copyright (c) 2005, Andrew J. Gryc   http://gryc.ws   andy@gryc.ws" << endl;
	cout << endl;
	cout << "Find/Replace Options" << endl;
	cout << "  -f : Finds strings only (DEFAULT)" << endl;
	cout << "  -r : Finds and replaces strings (must include <replace string> argument)" << endl;
	cout << "  -y : Case sensitive search and replace (DEFAULT)" << endl;
	cout << "  -i : Case insensitive search and replace" << endl;
	cout << "  -c : Case insensitive, replacement same case (lower/UPPER/Mixed) as original" << endl;
	cout << "  -u :*Unicode mode: search and replace are converted from ANSI to Unicode" << endl;
	cout << "  -n : Match if string is NOT found (cannot be used with -r)" << endl;
	cout << "  -o :*Ignore escape sequences in input strings" << endl;
	cout << "         (Example escapes that can be used: ^M \\r \\015 \\x0D)" << endl;
	cout << "  -w :*Match search string against entire line" << endl;
	cout << endl;
	cout << "Output Options" << endl;
	cout << "  -t : Print total summary" << endl;
	cout << "  -g : Print group summaries for each file and wildcard" << endl;
	cout << "  -m : Print matching lines (if replacing, prints after replacement)" << endl;
	cout << "  -p : Print matching lines (if replacing, prints original and replacement)" << endl;
	cout << "  -l : Print line number for matches" << endl;
	cout << "  -d : Print file and line number for matches" << endl;
	cout << "  -a : Use all common print options (equivalent to -tgm) (DEFAULT)" << endl;
	cout << endl;
	cout << "General Options" << endl;
	cout << "  -s :*Recursively process files in subdirectories" << endl;
	cout << "  -z :*Suppress printing file progress" << endl;
	cout << "  -k :*Create backups for replaced files" << endl;
	cout << "  -x :*Fully enumerate and sort file list before searching" << endl;
	cout << "         (Useful if replace is cyclically replacing in backup files)" << endl;
	cout << "  -b :*Force binary mode (auto-set if control chars in search/replace)" << endl;
	cout << "  -e :*Return count of matches as errorlevel to shell" << endl;
    cout << "  -j[<f>]*: Use index file f (defaults to .strfind.idx in current dir)" << endl;
    cout << "  -q :*Quick search; assumes index is up-to-date" << endl;
    cout << "  -v : Verbosity; -v=report times, -vv=index info, -vvv=debugging" << endl;
	cout << "  -h : Prints this help"<< endl;
    cout << endl;
    cout << "Options with * can be turned off by trailing dash (e.g. -k- is no backups)" << endl;
    cout << "Default options can be put into STRFIND_OPTIONS environment variable" << endl;
}

void ResetPrintOptions()
{
	static bool firstPrintArg = true;

	// We have some options turned on by default.
	// Because we don't allow turning options off, and we only
	// provide switches to turn things on, the very first time we see *any*
	// print option, reset all the rest of them.

	if (firstPrintArg)
	{
		PrintTotalStats		= false;
		PrintGroupStats		= false;
		PrintMatches		= false;
		PrintBeforeMatch	= false;
		PrintLinenumbers	= false;
		PrintFilenames		= false;

		firstPrintArg = false;
	}
}

void ProcessArg(const char *arg)
{
	while (*arg)
	{
		switch (toupper(arg[0]))
		{
        case 'V':
            Verbosity++;
            break;
		case 'F':
			PerformReplace = false;
			break;
		case 'R':
			PerformReplace = true;
			break;
		case 'Y':
			CaseInsensitive = false;
			break;
		case 'I':
			CaseInsensitive = true;
			break;
		case 'C':
			SameCapsForm = true;
			CaseInsensitive = true;
			break;
		case 'N':
			InvertedMatch = true;
			if (PerformReplace)
			{
				cerr << "Cannot use inverted match (-n) while doing replacement (-r)." << endl;
				exit(1);
			}
			break;
		case 'W':
			MatchWholeLine = (arg[1] != '-');  // Allow -w- to turn off option
            if (arg[1]=='-' || arg[1]=='+') arg++;
			break;
		case 'O':
			ProcessEscapes = (arg[1] == '-');  // Allow -o- to turn ON option
            if (arg[1]=='-' || arg[1]=='+') arg++;
			break;

		case 'T':
			ResetPrintOptions();
			PrintTotalStats = true;
			break;
		case 'G':
			ResetPrintOptions();
			PrintGroupStats = true;
			break;
		case 'M':
			ResetPrintOptions();
			PrintMatches = true;
			break;
		case 'P':
			ResetPrintOptions();
			PrintBeforeMatch = true;
			break;
		case 'L':
			ResetPrintOptions();
			PrintLinenumbers = true;
			break;
		case 'D':
			ResetPrintOptions();
			PrintFilenames = true;
			PrintLinenumbers = true;
			break;
		case 'A':
			ResetPrintOptions();
			PrintMatches = true;
			PrintLinenumbers = true;
			PrintTotalStats = true;
			break;

		case 'S':
			Subdirectories = (arg[1] != '-');  // Allow -s- to turn off option
            if (arg[1]=='-' || arg[1]=='+') arg++;
			break;
		case 'K':
			CreateBackups = (arg[1] != '-');  // Allow -k- to turn off option
            if (arg[1]=='-' || arg[1]=='+') arg++;
			break;
		case 'B':
			BinaryMode = (arg[1] != '-');  // Allow -b- to turn off option
            if (arg[1]=='-' || arg[1]=='+') arg++;
			break;
		case 'U':
			UnicodeMode	= (arg[1] != '-');  // Allow -u- to turn off option
            if (arg[1]=='-' || arg[1]=='+') arg++;
			break;
		case 'X':
			FullyEnumerateDirs = (arg[1] != '-');  // Allow -x- to turn off option
            if (arg[1]=='-' || arg[1]=='+') arg++;
			break;
		case 'Z':
			PrintProgress = (arg[1] == '-');  // Allow -z- to turn ON option
            if (arg[1]=='-' || arg[1]=='+') arg++;
			break;
		case 'E':
			ReturnErrorCount = (arg[1] != '-');  // Allow -e- to turn off option
            if (arg[1]=='-' || arg[1]=='+') arg++;
			break;
		case 'Q':
			QuickIndex = (arg[1] != '-');  // Allow -q- to turn off option
            if (arg[1]=='-' || arg[1]=='+') arg++;
			break;
        case 'J':
            // Allow -j- to turn off use of index file
            if (arg[1]=='-' && arg[2]==0)
                indexfilename[0] = 0;
            else if (arg[1])
                strcpy(indexfilename, arg+1);
            else
                strcpy(indexfilename, INDEX_FILE);
            return;

        case '?':
		case 'H':
			PrintHelp();
			return;

		default:
			cerr << "Unknown argument: " << arg << endl;
			exit(1);
		}

		// Go to next letter in the argument
		arg++;
	}
}

// Pads a string to the desired length with the requested char
string padto(const string &s, unsigned int length, char pad=' ')
{
	string printname(s);

	if (printname.length() > length)
		printname.erase(0, printname.length() - length);

	printname.append(length - printname.length(), pad);

	return printname;
}

// Print a bunch of data "debug style" with the given address
void BinaryPrint(streamoff addr, const string &line)
{
	const char *char_prefix = "  ";
	const unsigned int width= 80;
	const unsigned int maxchunk = 16;
	unsigned int p=0;
	unsigned int chunk; 
	unsigned int i;
	unsigned int remaining = line.length();

	while (remaining > 0)
	{
		// At beginning of line, print prefix
		cout << setw(8) << hex << setfill('0') << addr+p << " : ";
	
		// Figure out how much more we'll take for this line
		chunk = min(maxchunk, remaining);

		// Set default fill to 0
		cout << hex << setfill('0');

		for (i=p; i<p+chunk; i++)
		{
			cout << setw(2) << (unsigned int)(unsigned char)line.at(i) << " ";
		}

		// Spaces until the character block
		for (i=chunk; i<maxchunk; i++)
		{
			cout << "   ";
		}

		// Put corresponding characters
		cout << char_prefix;

		for (i=p; i<p+chunk; i++)
		{
			char c = line.at(i);
			if (c == '\r' || c=='\n' || c==0 || c=='\a' || c=='\t')
			{
				c = '.';
			} 

			cout << c;
		}

		cout << endl;

		p += chunk;
		remaining -= chunk;
	}

	cout << dec << endl;
}

// Print a line with any terminal movement characters replaced by a dot.
void PoliteLine(const string &line)
{
	for (string::const_iterator it=line.begin(); it!=line.end(); it++)
	{
		char c = *it;
	
		if (c == '\r' || c=='\n' || c=='\t')
		{
			cout << ' ';
		}
        else if (c=='\a')
        {
            cout << '.';
        }
		else if (c!=0)
		{
			cout << c;
		}
	}

	cout << endl;
}

void PrintMatchLine(const char *tag, const string &line, const string &filename, int linenumber)
{
	if (!PrintMatches || BinaryMode)  return;

	if (tag)
	{
		cout << tag << ": ";
	}

	if (PrintFilenames)
	{
		cout << filename ;
	} 

	if (PrintLinenumbers)
	{
		cout << "(" << linenumber << "): ";
	}

	PoliteLine(line);
}

const char *MatchesOrReplacements_msg()
{
	if (PerformReplace)
		return "replacement(s)";
	else
		return "match(es)";
}

// The control characters we recognize in search or replace strings
void ConvertControlChars(string &s)
{
	string create;

	for (const char *it = s.c_str(); *it; it++)
	{
		if (*it=='\\')
		{
			it++;

			switch (*it)
			{
			case 'a':
				create.append(1, 7); break;
			case 'b':
				create.append(1, 8); break;
			case 'n':
				create.append(1, 10); break;
			case 'v':
				create.append(1, 11); break;
			case 'r':
				create.append(1, 13); break;
			case 'f':
				create.append(1, 12); break;
			case '\\':
				create.append(1, '\\'); break;
			case 't':
				create.append(1, 9); break;
			case 'x':
				{
				it++;
				char *endchar;
				int val = strtoul(it, &endchar, 16);
				it = endchar - 1;
				create.append(1,val);
				}
				break;
			case '0':
			case '1':
			case '2':
			case '3':
				{
				char *endchar;
				it++;
				int val = strtoul(it, &endchar, 8);
				it = endchar - 1;
				create.append(1,val);
				}
				break;
			case '\'':
				create.append(1, '\''); break;
			case '\"':
				create.append(1, '\"'); break;
			default:
				cerr << "Invalid escape character (\\" << *it << ")" << endl;
				exit(2);
			}
		}
		else if (*it=='^')
		{
			it++;

			char c = toupper(*it);

			if (c<'@' || c>'_')
			{
				cerr << "Invalid escape character (^" << c << ")" << endl;
				exit(2);
			}
			create.append(1, c - '@');
		}
		else
		{
			create.append(1, *it);
		}
	}

	// Copy over our new string when done
	s = create;
}

bool string_equals(string &s1, string &s2)
{
	if (CaseInsensitive)
	{
		ci_string si1(s1.c_str());
		ci_string si2(s2.c_str());
		return si1==si2;
	}
	else
		return s1==s2;
}

int string_find(string &containing, string &sought, int pos)
{
	if (CaseInsensitive)
	{
		ci_string ci_containing(containing.c_str());
		ci_string ci_sought(sought.c_str());
		return ci_containing.find(ci_sought,pos);
	}
	else
		return containing.find(sought,pos);
}

int string_case_consistent(string &line, int pos, int strlen)
{
	bool all_upper = true;
	bool all_lower = true;
	
	for (int i=pos; i<pos+strlen; i++)
	{
		char ch = line[i];
		
		if (ch!=toupper(ch))
			all_upper = false;

		if (ch!=tolower(ch))
			all_lower = false;
	}

	if (all_upper)
		return 1;
	else if (all_lower)
		return -1;
	else
		return 0;
}

string string_lower(string &str)
{
	string s(str);

	transform(s.begin(), s.end(), s.begin(), (int(*)(int)) tolower);
	return s;
}

string string_upper(string &str)
{
	string s(str);

	transform(s.begin(), s.end(), s.begin(), (int(*)(int)) toupper);
	return s;
}

void string_replace(string &line, int pos, int strlen, string &replace)
{
	if (SameCapsForm)
	{
		int test = string_case_consistent(line, pos, strlen);

		string CapsReplacement;

		switch (test)
		{
		case 1:
			CapsReplacement = string_upper(replace);
			break;
		case -1:
			CapsReplacement = string_lower(replace);
			break;
		case 0:
			CapsReplacement = replace;
			break;
		}

		line.replace(pos, strlen, CapsReplacement);
	}
	else
	{
		line.replace(pos, strlen, replace);
	}
}

void ShowProgress(const char *filename, streamoff pos, streamoff len)
{
	// Print progress, trim to 80 characters, and possibly show a percentage if we
	// spend too long on one file...
	if (PrintProgress)
	{
		string name(filename);

		if (len!=0)
		{
			int percent = (int)((pos*100.0)/len);
			char buffer[20];

			itoa(percent, buffer, 10);

			name += " (";
			name += buffer;
			name += "%)";
		}
		fprintf(stderr,"%s\r", padto(name,MAX_DISPLAYED_CHARS).c_str());

		showing_progress = true;
	}
}

void EraseProgress()
{
	if (showing_progress)
	{
		// Erase the filename we printed before we show the match so there aren't any remnants
		if (PrintProgress)  ShowProgress("",0,0);
		showing_progress = false;
	}
}

int ReplaceStrings(const char *filename, string searchStr, string replaceStr)
{
	const char *TempFilename = tempnam(NULL,NULL);
	const char *output_file = (PerformReplace)?TempFilename:NULL_DEVICE;

	// Used to set ios::binary if either in BinaryMode or UnicodeMode, but leave text otherwise.  
	// Now we always do it, since there appear to be bugs in the text mode implementation
	// for VC++ under certain circumstances.  The problem specifically seems to be
	// strings of characters that confuse the line parser, reset the buffered input, 
	// and will cause loops in the reading.  You'll keep redoing the same couple lines...
	// (See comments for the function input_line.)
	int binary_flag = ios::binary;	

    // If we are using an index, we may be able to shorten the search
    if (idx)
    {
        BitsetType testbits;

        // The only time we completely avoid the search is if 
        // the file in question is in the index, and the index file is newer
        if (idx->FindFile(filename, testbits) && !idx->NewerThanIndex(filename))
        {
            if (!idx->PossibleMatch(searchStr.c_str(), testbits))
                return 0;
        }
    }

    // If we are using an index, we may be able to shorten the search
    BitsetType searchbits;

	// Show that we started searching this file
	ShowProgress(filename, 0, 0);

    // Open in and out streams
	ifstream input(filename, ios::in|binary_flag);
	ofstream output(output_file, ios::out|binary_flag);

	input.seekg(0, ios::end);
	streamoff filelen = input.tellg();
	input.seekg(0, ios::beg);

	if (!input.is_open())
	{
		cerr << "Error opening \"" << filename << "\" for input." << endl;
		return 0;
	}

	if (!output.is_open())
	{
		cerr << "Error opening \"" << output_file << "\" for output." << endl;
		return 0;
	}

	int count = 0;
    bool printheader = true;
	int linenumber = 0;
	string line;

	DWORD tick = GetTickCount(); 

	while (!input.eof())
	{
		streamoff oldfilepos = input.tellg();
		streamoff newfilepos = output.tellp();

		// For very big files, make sure we show some progress as we crank through...
		if (PrintProgress)
		{
			if (GetTickCount() - tick > 1000)
			{
				tick = GetTickCount();
				ShowProgress(filename, oldfilepos, filelen);
			}
		}

		input_line(input, line);

		// Only process it if we were successful
		if (!input.bad())
		{
			linenumber++;


            // If indexing, add line to index
            if (idx)
            {
                idx->AddBitset(searchbits, line.c_str());
            }

			// Save a copy of the original for printing before matches
			string original = line;
			bool matches = false;
			int matchesthisline = 0;

			// Do finds for the whole line
			if (MatchWholeLine)
			{
				if (string_equals(line, searchStr))
				{
					matchesthisline++;
					matches=true;

					if (PrintMatches && BinaryMode)
					{
						EraseProgress();

                        // Print header before first match in file
                        if (PrintGroupStats && printheader)
                        {
			                cout << MATCHHEADER << filename << endl;
                            printheader = false;
                        }

						if (PerformReplace)
							BinaryPrint(newfilepos, replaceStr);
						else
							BinaryPrint(oldfilepos, searchStr);
					}

					if (PerformReplace)
					{
						string_replace(line, 0, line.length(), replaceStr);
					}
				}
			}
			else	// Do finds within a line
			{
				int pos = 0;
				do 
				{
					pos = string_find(line, searchStr, pos);

					if (pos != string::npos)
					{
						if (PrintMatches && BinaryMode)
						{
							EraseProgress();

                            // Print header before first match in file
                            if (PrintGroupStats && printheader)
                            {
			                    cout << MATCHHEADER << filename << endl;
                                printheader = false;
                            }

                            if (PerformReplace)
								BinaryPrint(newfilepos+pos, replaceStr);
							else
								BinaryPrint(oldfilepos+pos, searchStr);
						}

						if (PerformReplace)
						{
							string_replace(line, pos, searchStr.length(), replaceStr);
							pos += replaceStr.length();
						}
						else 
						{
							pos += searchStr.length(); 
						}
						matchesthisline++;
						matches = true;
					}

				} while (pos != string::npos);
			}

			// Process negative matches
			if (InvertedMatch)
			{
				matches = !matches;
				if (matches) matchesthisline = 1;
			}

			// Add up matches on this line
			if (matches) count += matchesthisline;

			// Output any requested printing information
			if (matches)
			{
				EraseProgress();

                // Print header before first match in file
                if (PrintGroupStats && printheader)
                {
			        cout << MATCHHEADER << filename << endl;
                    printheader = false;
                }

				if (PerformReplace && PrintBeforeMatch)
				{
					PrintMatchLine("Original", original, filename, linenumber);
					PrintMatchLine("REPLACED", line, filename, linenumber);
				}
				else
				{
					PrintMatchLine(0, line, filename, linenumber);
				}
			}

			// Output the replaced line to the temporary file.
			if (PerformReplace)
			{
				output << line;
			}
		}
	}

	output.close();

	// Only if there are any replacements, create backup and overwrite the old file
	if (count!=0) 
	{
		if (PerformReplace)
		{
			if (CreateBackups)
			{
				string BackupFilename(filename);
				BackupFilename.append(".bak");
				CopyFile(filename, BackupFilename.c_str(), false);
			}

			// Copy the replacement file over the original
			CopyFile(TempFilename, filename, false);
		}

		// Print the summary for this file
		if (PrintGroupStats)
		{
			EraseProgress();

			cout << MATCHSTRING << filename << ": " << count << " " << MatchesOrReplacements_msg() << endl;

			// Put in a blank line for readability if we're outputting file contents too
			if (PrintMatches) cout << endl;
		}
	}

	// Remove the temporary file
	DeleteFile(TempFilename);

    // If indexing, save created index
    if (idx)
    {
        idx->AddToIndex(filename, searchbits);
        idx->PeriodicSave();
    }

	return count;
}

void SplitDirectoryFromWildcard(const string &wildcard, string &base_directory, string &trailing_wildcard)
{
	base_directory = wildcard;
	int slash = base_directory.rfind(DIR_SEPARATOR);

	if (slash==string::npos)
	{
		// If no backslash, set the directory to dot ("current"), and use the wildcard for files.
		base_directory=".";
		trailing_wildcard = wildcard;
	}
	else
	{
		// For recursed subdirs, get the filename part of the wildcard
		// and erase it off the base directory.
		trailing_wildcard.assign(wildcard.c_str() + slash + 1);
		base_directory.erase(slash);
	}
}


int GetAllFiles(vector<string> &file_list, const string &recursed_dirs, const string &wildcard, int *count, const string &search, const string &replace)
{
	WIN32_FIND_DATA FindData;
	HANDLE hFindHandle;
	int total_files = 0;

	string fullpath(recursed_dirs);
	fullpath += DIR_SEPARATOR;
	fullpath += wildcard;

	hFindHandle = FindFirstFile(fullpath.c_str(), &FindData);

	if (hFindHandle != INVALID_HANDLE_VALUE)
	{
		// Process all normal files first
		do
		{
			if (!(FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
			{
                // Do not search index files
                if (!strcmp(INDEX_FILE, FindData.cFileName))
                    continue;

				string tempfilename(recursed_dirs);
				
				if (tempfilename.length() != 0)
				{
					tempfilename += DIR_SEPARATOR;
				}

				tempfilename += FindData.cFileName;

                // Do the work
                if (FullyEnumerateDirs)
					file_list.push_back(tempfilename);
				else 
					*count += ReplaceStrings(tempfilename.c_str(), search, replace);

				total_files++;
			}
		} while (FindNextFile(hFindHandle, &FindData));

		FindClose(hFindHandle);
	}

	// If requested, modify the request to do subdirs, and do them now.
	// We don't do this in the normal loop because if they supply a
	// wildcard, we want it to match all directories.
	if (Subdirectories)
	{
		string dir_wildcard = recursed_dirs + DIR_SEPARATOR + "*.*";

		hFindHandle = FindFirstFile(dir_wildcard.c_str(), &FindData);

		if (hFindHandle == INVALID_HANDLE_VALUE)
		{
			return total_files;
		}

		// Process all subdirectories
		do
		{
			if (FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
			{
				// Don't process . or ..
				if (FindData.cFileName[0]!='.')
				{
					string tempdirectory(recursed_dirs);
					tempdirectory += DIR_SEPARATOR;
					tempdirectory += FindData.cFileName;

					string tempwildcard(tempdirectory);
					tempwildcard += DIR_SEPARATOR;
					tempwildcard += wildcard;

					ShowProgress(tempwildcard.c_str(), 0, 0);

					total_files += GetAllFiles(file_list, tempdirectory, wildcard, count, search, replace);
				}
			}
		} while (FindNextFile(hFindHandle, &FindData));

		FindClose(hFindHandle);
	}

	return total_files;
}

int SearchFileList(vector<string> &file_list,const string &search, const string &replace)
{
    int count = 0;

	// Now, actually process each file.
	// For politeness, we sort the list.
	sort(file_list.begin(), file_list.end());

	for (vector<string>::iterator it = file_list.begin(); it!=file_list.end(); it++)
	{
		count += ReplaceStrings(it->c_str(), search, replace);
	}
    return count;
}

bool MatchesWildcard(const char *wildcard, const char *filename)
{
    // Rather simplistic algorithm: if the first char is a star,
    // just match the extension; else match the whole thing.
    // Note that we don't bother with "?" or anything else more
    // sophisticated.
    if (wildcard[0]=='*')
        return !stricmp(wildcard+1, filename+strlen(filename)-strlen(wildcard+1));
    else
        return !stricmp(wildcard, filename);
}

int GetMatchingIndexedFiles(vector<string> &file_list, const string &base_directory, const string &trailing_wildcard, const string &search)
{
    int count = 0;
    BitsetType searchbits;

    idx->AddBitset(searchbits, search.c_str());

    for (IndexFile::IndexIterator it=idx->begin(); it!=idx->end(); it++)
    {
        // Test filename wildcard against index name, and only do this if a match
        if (MatchesWildcard(trailing_wildcard.c_str(), it->first.c_str()))
        {
            if (idx->FindBits(searchbits, it))
            {
			    ShowProgress(it->first.c_str(), 0, 0);
                file_list.push_back(it->first);
            }

            count++;
        }
    }

    return count;
}

int MultipleFileReplaceStrings(const char *wildcard, const string &search, const string &replace)
{
	// First construct the list of files we're to replace.
	// We do this in two stages like this because for wildcard arguments
	// (*.*) we'd otherwise process backup files and the temporary file
	// we create during processing.  Both of them can cause us to 
	// enter an endless loop of creating backups or adding to the temp
	// file.

	int count = 0;
	int total_files = 0;
	vector<string> file_list;

	// Have to separate argument into directory and wildards
	string base_directory;
	string trailing_wildcard;

	SplitDirectoryFromWildcard(wildcard, base_directory, trailing_wildcard);

    if (QuickIndex && idx)
    {
        total_files = GetMatchingIndexedFiles(file_list, base_directory, trailing_wildcard, search);
        
        count = SearchFileList(file_list, search, replace);
    }
    else if (FullyEnumerateDirs)
	{
		total_files = GetAllFiles(file_list, base_directory, trailing_wildcard, NULL, search, replace);
		
        count = SearchFileList(file_list, search, replace);
	}
	else
	{
		total_files = GetAllFiles(file_list, base_directory, trailing_wildcard, &count, search, replace);
	}

	// And print summary information for this particular wildcard group
	if (PrintGroupStats)
	{
		EraseProgress();

		cout << MATCHSTRING << wildcard << ": " << count << " " << MatchesOrReplacements_msg()
			 <<	" in " << total_files <<  " file(s)" << endl;

		// Put in a blank line for readability if we're outputting file contents too
		if (PrintMatches) cout << endl;
	}

	return count;
}

bool WarnBinaryModeSearch(const string &str)
{
	// Look for an embedded \r or \n.  That's what we use as a line terminator
	// in binary mode, and we want to warn them if their search string contains
	// them in the middle that we may not find the pattern broken across lines.

	unsigned int len = str.length();
	bool found_embedded = false;

	// The combinations we use for line termination are:
	//		\r\n, \r, or \n.
	// So, we can find a \r at the last or second to last char,
	// or a \n at the last character, but not anywhere else.

	for (unsigned int i=0; i<len; i++)
	{
		if (str.at(i)=='\n' && i<len-1)
			found_embedded = true;

		if (str.at(i)=='\r' && i<len-2)
			found_embedded = true;
	}

	return found_embedded;
}

void ConvertToUnicode(string &str)
{
	string::iterator it=str.begin(); 
	
	while (it != str.end())
	{
		it++;
		it = str.insert(it, 0);
		it++;
	}
}

int main(int argc, char* argv[])
{
    DWORD starttime = GetTickCount();

	if (argc <= 1)
	{
		PrintHelp();
		exit(0);
	}

    // Process default environment variable arguments first
    char *env = getenv("STRFIND_OPTIONS");
    if (env)
    {
       const char * token = strtok(env, " \t");
       
       while (token)
       {
          if (token[0]=='-' || token[0]=='/')
              ProcessArg(token+1);
          else
              cerr << "STRFIND_OPTIONS environment variable cannot contain non-options (" << token << ")" << endl;

          token = strtok(0, " \t");
       }
    }

	// Now process command line arguments
	int startarg=1;

	while ((startarg < argc) && (argv[startarg][0]=='-' || argv[startarg][0]=='/'))
	{
		ProcessArg(argv[startarg]+1);

		startarg++;
	}

	// Get search string
	if (startarg >= argc) {
		cerr << "No search string argument." << endl;
		exit(1);
	}

    const char *search = argv[startarg++];

    const char *replace;

	// Get replace string (if any)
	if (PerformReplace)
	{
		if (startarg >= argc) {
			cerr << "No replacement string argument." << endl;
			exit(1);
		}

		replace = argv[startarg++];
	}
	else
		replace = search;

	// Convert the strings to escaped versions
	string searchStr = search;
	string replaceStr = replace;

	bool ContainsEscapes = (searchStr.find("\\")!=string::npos || 
							replaceStr.find("\\")!=string::npos||
							searchStr.find("^")!=string::npos  || 
							replaceStr.find("^")!=string::npos);

	if (ContainsEscapes)
	{
		if (ProcessEscapes)
		{
			ConvertControlChars(searchStr);
			ConvertControlChars(replaceStr);
			BinaryMode = true;
		}
		else
		{
			cout << "Verbatim mode: \\ and ^ processing not performed on search/replace strings." << endl;
		}
	}

	// If unicode mode, convert ANSI strings into Unicode equivalents.
	// Note that this doesn't force us into binary, like it would if we 
	// passed in strings with alternating \000's in them.
	if (UnicodeMode)
	{
		ConvertToUnicode(searchStr);
		ConvertToUnicode(replaceStr);
	}

	// Check for line matches that may not work properly and exit
	if (WarnBinaryModeSearch(searchStr))
	{
		cerr << "Error: \\r or \\n found in middle of search string.  StrFind cannot find strings" << endl;
		cerr << "that span multiple lines, even in binary mode." << endl;
		exit(3);
	}

    // If index in use, load it now
    if (indexfilename[0])
    {
        idx = new IndexFile(indexfilename);

        idx->LoadFile();
    }

	// Process file specifications
	int count = 0;
	while (startarg < argc)
	{
		const char *wildcard = argv[startarg++];
		count += MultipleFileReplaceStrings(wildcard, searchStr, replaceStr);
	}

	// Print results
	if (PrintTotalStats)
	{
		const char *match_item_msg = (MatchWholeLine) ? "line(s)" : "instance(s)";

		const char *match_type_msg = (InvertedMatch) ? "not found" : "found";


		if (PerformReplace)
			cout << ">> Total: " << count << " " << match_item_msg << " of \"" << search <<  
				    "\" replaced with \"" << replace << "\"." << endl;
		else
			cout << ">> Total: " << count << " " << match_item_msg << " of \"" << search <<  
				    "\" " << match_type_msg << "." << endl;

        DWORD endtime = GetTickCount();
        float secs = (endtime-starttime)*0.001f;

        if (Verbosity >= 1)
        {
            cout << endl << "Total time = " << secs << " seconds" << endl;
        }
    }

    // Save index
    if (idx)
    {
        idx->SaveFile();
        delete idx;
        idx = 0;
    }

	if (ReturnErrorCount)
		return count;
	else
		return 0;
}
